from pathlib import Path
import json
from functools import partial
from typing import Dict, Union, List, Tuple, Any
import random


def get_ner_data_class():
    try:
        from torch.utils.data.dataset import Dataset
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Please install pytorch first https://pytorch.org/")

    class NERDataset(Dataset):
        """
        NER Dataset
        return text, a list of tags
        """

        def __init__(self, data):
            self.data = data
            self.labels = []
            for index, labels in data['labels'].items():
                self.labels += list(filter(
                    lambda x: "skipped" not in x, labels))
            self.texts = data['texts']
            self.options = data['options']
            self.i2c = dict(enumerate(["O", ]+self.options))
            self.c2i = dict((v, k) for k, v in enumerate(["O", ]+self.options))
            self.passon_kwargs = dict()

        def __repr__(self):
            options = self.data["options"]
            return f"NERDataset, {options}\n{len(self)}rows"

        def __len__(self):
            return len(self.labels)

        def __getitem__(self, idx):
            label = self.labels[idx]
            text = self.texts[str(label["pandas"])]
            return text, label['tags']

        def split_train_valid(self, valid_ratio: float = .2):
            """
            split dataset to train and valid.
            since one text can have multiple tags
                 the valid_ratio will be applied to text
                 instead of labels

            """
            all_train_index = []
            train_texts, val_texts = dict(), dict()
            train_labels, val_labels = dict(), dict()

            # when spliting data we put all the textinto
            #   both train and valid
            for index, text in self.texts.items():
                index = str(index)
                if random.random() > valid_ratio:
                    all_train_index.append(str(index))
                train_texts.update({index: text})
                val_texts.update({index: text})

            # we only separate the train valid for labels
            for idx, label_list in self.data["labels"].items():
                index = str(label_list[0]["pandas"])
                if index in all_train_index:
                    train_labels.update({index: label_list})
                else:
                    val_labels.update({index: label_list})

            train_data, val_data = dict(), dict()

            for k, v in self.data.items():
                if k == "texts":
                    train_data.update({"texts": train_texts})
                    val_data.update({"texts": val_texts})
                elif k == "labels":
                    train_data.update({"labels": train_labels})
                    val_data.update({"labels": val_labels})
                else:
                    train_data.update({k: v})
                    val_data.update({k: v})
            # instantiate by the same class
            return self.__class__(train_data, **self.passon_kwargs),\
                self.__class__(val_data, **self.passon_kwargs)
    return NERDataset


def get_ner_hf_class():
    try:
        from torch.utils.data.dataloader import DataLoader
        import torch
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Please install pytorch first https://pytorch.org/")
    NERDataset = get_ner_data_class()

    class NERDatasetHF(NERDataset):
        """
        NER Dataset of HuggingFace tokenized sentences
        output Tuple:
            x, y
        """

        def __init__(
            self,
            data: Dict[str, Any],
            tokenizer,
            tokenization_options: Dict[str, Any] = dict({
                'return_offsets_mapping': True,
                'padding': 'max_length',
                'return_tensors': 'pt',
                'truncation': True,
                'max_length': 512}),
        ):
            """
            data: contains at least the following keys
                texts: Dict[str, Any], a dictionary of texts, could be empty
                labels: Dict[str, Any], a dictionary of labels, could be empty
                options: Dict[str, Any], a dictionary of options,
                    could be empty

            tokenizer: a huggingface tokenizer class object

            tokenization_options: tokenization options for tokenizer
                eg.{'return_offsets_mapping': True,
                    'padding': 'max_length',
                    'return_tensors': 'pt',
                    'truncation': True,
                    'max_length': 512}
            """
            super().__init__(data)
            self.options = self.data["options"]

            self.tokenizer = tokenizer

            # default tokenization options
            # this will be overriden by the tokenization_options
            tkkw = dict(
                return_offsets_mapping=True,
                padding=True,
                return_tensors="pt",
            )

            tkkw.update(tokenization_options)
            self.tokenization_options = tkkw
            self.tokenizing = partial(
                self.tokenizer,
                **tkkw
            )

            # keyword arguments passed on to inherited class
            self.passon_kwargs = dict(
                {
                    "tokenizer": tokenizer,
                    "tokenization_options": tokenization_options}
            )

        def __repr__(self):
            return f"NER dataset (HuggingFace Tokenizer):\n\t{len(self)} rows"

        @staticmethod
        def tag_to_offset(tag: Dict[str, Union[str, int]]):
            offset = tag['offset']
            return [offset, len(tag['text'])+offset], tag['label']

        def collate(self, batch):
            texts, labels = zip(*batch)
            tked = self.tokenizing(list(texts))
            x = tked['input_ids']
            offset_mapping = tked['offset_mapping']

            # mark label from an all zero tensor
            y = torch.zeros_like(x)

            for row in range(x.size(0)):
                tags_pos = list(map(self.tag_to_offset, labels[row]))

                for tag_pos, tag_label in tags_pos:

                    tag_pos = torch.LongTensor(tag_pos)
                    tag_mask = (tag_pos[0] <= offset_mapping[row, :, 0]) * \
                        (tag_pos[1] >= offset_mapping[row, :, 1]) * \
                        self.c2i[tag_label]
#                     text_recorver = texts[row][tag_pos[0]: tag_pos[1]]
#                     print(f"text reslice:{text_recorver}")
#                     token_decoded = self.tokenizer.decode(
#                        list(x[row, (tag_mask !=0 )]))
#                     print(f"token decoded: {token_decoded}")

                    y[row, :len(tag_mask)] += tag_mask
            return {
                "input_ids": x,
                "attention_mask": tked["attention_mask"],
                "targets": y,
                "offset_mapping": offset_mapping,
            }

        def get_data_loader(self, **kwargs):
            """
            input:
                - batch_size
                - shuffle
                - num_workers
                - ... other key word arguments for dataloader
                    except collate_fn, which will be default
                    to self.collate
            """
            return DataLoader(self, collate_fn=self.collate, **kwargs)

        def one_batch(
            self, batch_size: int = 2
        ) -> Tuple[torch.LongTensor]:
            """
            sample 1 batch of data
            """
            return next(iter(
                self.get_data_loader(
                    batch_size=batch_size, shuffle=True)))

        @staticmethod
        def tensor_position_map(
            x: torch.LongTensor
        ) -> torch.LongTensor:
            idx_by_row = torch.arange(
                x.size(0))[:, None].repeat([1, x.size(1)])
            idx_by_tk = torch.arange(x.size(1))[None, :].repeat([x.size(0), 1])

            row_tk = torch.cat(
                [idx_by_row[:, :, None], idx_by_tk[:, :, None]], dim=-1)
            return row_tk

        @staticmethod
        def equal_to_next_map(
            y: torch.LongTensor
        ) -> torch.BoolTensor:
            equals_next = y[:, :-1] == y[:, 1:]
            equals_next = torch.cat([
                equals_next,
                torch.zeros((equals_next.size(0), 1)).bool()],
                dim=1)
            return equals_next

        def decode(
            self, batch: Dict[str, torch.Tensor], y: torch.Tensor,
        ) -> List[Dict[str, Union[int, str]]]:
            """
            convert batch input, y_index tensor
                back to entity metadata
            """
            x = batch["input_ids"]
            offsets = batch["offset_mapping"]

            equals_next = self.equal_to_next_map(y)
            mapping = self.tensor_position_map(x)
            # has positive prediction
            has_pos = (y != 0)

            ids = []
            tokens = []
            start_pos = None

            for e, offset, x_1, y_1, (row_id, token_id) in zip(
                    equals_next[has_pos],
                    offsets[has_pos],
                    x[has_pos], y[has_pos], mapping[has_pos]):

                ids.append(x_1.item())
                if start_pos is None:
                    start_pos = offset[0]
                if e.item() is False:
                    tokens.append({
                        "row_id": row_id.item(),
                        "token_id": token_id.item(),
                        "text": self.tokenizer.decode(ids),
                        "label": self.i2c[y_1.item()],
                        "offset": int(start_pos.item()),
                    })
                    ids = []
                    start_pos = None

            return tokens
    return NERDatasetHF


def load_ner_data_pytorch(file_path: Path):
    """
    Automate the process from langhuan export
        to a pytorch dataloader
    """
    with open(file_path, "r") as f:
        json_data = json.loads(f.read())

    NERDataset = get_ner_data_class()

    return NERDataset(json_data)


def load_ner_data_pytorch_huggingface(
    file_path: Path,
    tokenizer,
    tokenization_options: Dict[str, Any] = dict({
                'return_offsets_mapping': True,
                'padding': 'max_length',
                'return_tensors': 'pt',
                'truncation': True,
                'max_length': 512}),
):
    """
    Automate the process from langhuan export
        to a pytorch dataloader,
    Outputing long tensors tokenized by a
        huggingface tokenizer
    """
    with open(file_path, "r") as f:
        json_data = json.loads(f.read())

    NERDatasetHF = get_ner_hf_class()

    return NERDatasetHF(json_data, tokenizer=tokenizer,
                        tokenization_options=tokenization_options, )



from pathlib import Path
from .utility import now_str
import os
import json
import logging
from typing import Dict
import pandas as pd
from datetime import datetime
from uuid import uuid4


def now():
    return datetime.now().strftime("%Y%m%d_%H%M%S")


def openj(file):
    with open(file, "r") as f:
        return json.loads(f.read())


def history_from_json(file):
    """
    read history from a single JSON file
    """
    df = pd.DataFrame(openj(file))
    df["fname"] = Path(file).name
    return df


def to_one_line(df):
    """
    Combine the duplicate tag into 1 line
    """
    if len(df) == 1:
        return df
    skipped = df["skipped"].sum()

    # filter out the skipped as options
    if (skipped > 0) and (skipped < len(df)):
        df = df[df.skipped.isna()]
    return df.sort_values(by="now", ascending=False).head(1)


def clean_history(history_df):
    history_df["index"] = history_df["index"].apply(int)
    newdf = history_df.groupby(["pandas", "user_id"])\
        .apply(to_one_line)
    newdf = newdf\
        .reset_index(drop=True)

    if "level_0" in newdf:
        newdf = newdf.drop("level_0", axis=1)
    if "level_2" in newdf:
        newdf = newdf.drop("level_2", axis=1)
    return newdf


def combine_history(HISTORY: Path) -> Path:
    """
    combine multiple history json files into 1 file
        remove the duplicate by user_id and pandas
    """
    HISTORY_FILES = list(
        i for i in HISTORY.iterdir() if i.suffix == ".json")

    # concatenate all the history files into 1 single dataframe
    history_df = pd.concat(list(map(history_from_json, HISTORY_FILES)))\
        .reset_index(drop=True)
    if "level_2" in history_df:
        history_df = history_df.drop(
            "level_2", axis=1)
    history_df = clean_history(history_df)
    new_history = HISTORY/f"history_combined_{now()}.json"
    history_df.to_json(new_history, orient="records")
    logging.info(f"save to {new_history}")
    for i in HISTORY_FILES:
        logging.info(f"rm -f {i}")
        os.system(f"rm -f {i}")
    return new_history


class History:
    """
    History log handler
    """

    def __init__(
        self,
        task_name: str,
        load_history: bool = False,
        save_frequency: int = 42,
    ):
        """
        task_name: str, name of the task
            if $HOME/.cache/langhuan/{task_name} exist
        """
        self.task_name = task_name
        self.load_history = load_history
        self.save_frequency = save_frequency
        self.home = Path(os.environ["HOME"])
        self.history = []
        self.history_save_mark = 0
        self.batcher = 0

        self.cache = self.home / ".cache"
        self.cache.mkdir(exist_ok=True)

        self.langhuan = self.cache / "langhuan"
        self.langhuan.mkdir(exist_ok=True)

        self.task = self.langhuan / task_name

        if self.load_history:
            if self.task.exists():
                self.history = self.read_all_history()
            else:
                logging.error(
                    f"no history under {self.task}")
                logging.error(
                    "you can try changing the task_name"
                )
                tasks = os.listdir(self.langhuan)
                logging.error(
                    f"valid task names:\n{tasks}"
                )
        self.task.mkdir(exist_ok=True)

    def __repr__(self):
        return f"CacheFolder:\t{self.task}"

    def save_new_history(self):
        logging.info(f"current history length: {len(self.history)}")
        logging.info(f"current save mark: {self.history_save_mark}")
        to_save = self.history[self.history_save_mark:]
        self.history_save_mark += len(to_save)

        hexstr = str(uuid4())[-6:]
        with open(self.task/f"history_{now_str()}_{hexstr}.json", "w") as f:
            f.write(json.dumps(to_save))

    def read_all_history(self):
        """
        read the saved history from self.task
        """
        history_data = []
        history_path = combine_history(self.task)
        with open(history_path, "r") as f:
            history_data = json.loads(f.read())

        self.history_save_mark = len(history_data)

        return history_data

    def __add__(self, x: Dict[str, str]):
        """
        Entrying new data
        """
        self.history.append(x)

        # batcher is used to save the history in batches
        self.batcher += 1
        if self.batcher % self.save_frequency == 0:
            logging.info(f"auto save triggered: {self.batcher}")
            self.save_new_history()
        return self.history



from abc import ABC
from typing import List


def check_column_set(column: str) -> None:
    if column is None:
        raise KeyError(
            "you have to set 'order_by_column' " +
            "when using this strategy, " +
            "preferably a score between 0 ~ 1"
        )


class OrderStrategies(ABC):
    """
    Some known strategies are available(as you can just use
    the name of the strategy), start and end are ordered
    according to the value of order_by_column:
    - forward_march:
        honest to the earth start to end
        the simplest ordering strategy
        The default is in ascending order
        If you want descending order, you can design
        your order by column as -1 x your order by column
    - pincer: move 1 by 1 both from start and from end.
    - trident: move 1 by 1 from start, from end and from
        the middle. It's like 2 pincer
    """

    def mix_streams(self, *streams):
        if len(streams) == 0:
            return []
        elif len(streams) == 1:
            return list(streams[0])

        min_length = min(list(map(len, streams)))
        combined = []
        for i in range(min_length):
            for stream in streams:
                combined.append(stream[i])

        combined += self.mix_streams(*list(
            stream[min_length:] for stream in streams
            if len(stream[min_length:]) > 0))

        return combined

    def forward_march(self, **kwargs) -> List[int]:
        """
        Simplist order strategy
        move 1 by 1 both from start and from end.
        """
        return list(self.df.index)

    def pincer(self, **kwargs) -> List[int]:
        order_by_column = kwargs.get("order_by_column")
        check_column_set(order_by_column)

        ordered_idx = list(
            self.df.sort_values(by=[order_by_column, ]).index)

        mid_point = len(ordered_idx)//2
        return self.mix_streams(
            ordered_idx[:mid_point], ordered_idx[mid_point:][::-1])

    def trident(
        self,
        **kwargs
    ) -> List[int]:
        order_by_column = kwargs.get("order_by_column")
        check_column_set(order_by_column)

        mid_score = self.df[order_by_column].min() +\
            (self.df[order_by_column].max() -
             self.df[order_by_column].min())/2

        sorted_df = self.df.sort_values(by=[order_by_column, ])

        bigger = sorted_df.query(f"{order_by_column}>={mid_score}")
        smaller = sorted_df.query(f"{order_by_column}<{mid_score}")

        bigger_mid_point = len(bigger)//2
        smaller_mid_point = len(smaller)//2

        return self.mix_streams(
            list(bigger.index)[:bigger_mid_point],
            list(bigger.index)[bigger_mid_point:][::-1],
            list(smaller.index)[:smaller_mid_point],
            list(smaller.index)[smaller_mid_point:][::-1],
        )



import pandas as pd
from itertools import chain


class Options(object):
    def __init__(
            self,
            df: pd.DataFrame, options,):
        if options is None:
            options = []
        # when options => a list of options
        if type(options) in [list, set]:
            self.option_vals = [list(options), ] * len(df)
            self.known_options = list(set(options))

        # when options => a name of df column
        elif type(options) == str:
            assert options in df,\
                f"when options set to string, it has to be one of {df.columns}"

            self.option_vals = df[options]
            self.known_options = self.calc_known_options(self.option_vals)
        else:
            raise TypeError(
                f"""
        options type has to be one of 'list, str',
        yours: {type(options)}
        """)

        self.df_idx = df.index

        self.df = pd.DataFrame(
            dict(options=self.option_vals, idx=self.df_idx))

        self.option_col = self.df["options"]
        self.df = self.df.set_index("idx")

    def __len__(self): return len(self.df)

    def __repr__(self):
        return f"options:{self.known_options}"

    @staticmethod
    def calc_known_options(iterable):
        return list(set(list(chain(*list(iterable)))))

    def __getitem__(self, idx):
        options = dict(self.df.loc[idx])["options"]
        return options

    def new_for_option_col(self, option):
        def new_for_option_col_(x):
            if option in x:
                return x
            else:
                x.append(option)
                return x
        return new_for_option_col_

    def delete_for_option_col(self, option):
        def delete_for_option_col_(x):
            if option in x:
                x.remove(option)
                return x
            else:
                return x
        return delete_for_option_col_

    def add_option(self, option: str):
        if option not in self.known_options:
            self.known_options.append(option)

        self.option_col = self.option_col.apply(
            self.new_for_option_col(option)
        )

        return self.known_options

    def delete_option(self, option: str):
        if option in self.known_options:
            self.known_options.remove(option)
        self.option_col = self.option_col.apply(
            self.delete_for_option_col(option)
        )

        return self.known_options



from datetime import datetime
from flask import request
from typing import Union
import json
from pathlib import Path
import regex


def now_str(): return datetime.now().strftime("%y%m%d_%H%M%S")


def now_specific():
    return datetime.now().strftime("%y-%m-%d_%H:%M:%S")


def cleanup_tags(x: str) -> str:
    """
    remove the string that will break the frontend
    x: str, input string
    """
    return x.replace("<", "˂").replace(">", "˃")


def arg_by_key(key: str) -> Union[str, int, float]:
    """
    get a value either by GET or POST method
    with api function
    """
    if request.method == "POST":
        data = json.loads(request.data)
        rt = data[key]
    elif request.method == "GET":
        rt = request.args.get(key)
    else:
        raise RuntimeError(
            "method has to be GET or POST")
    return rt


def get_root() -> Path:
    return Path(__file__).parent.absolute()


def findall_word_position(text: str, word: str) -> list:
    """
    find all the position of word in text
    """
    text = text.lower()
    word = word.lower()
    return [m.start() for m in regex.finditer(word, text)]
